#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import re
import string
import sys
from typing import Dict, Iterable, Sequence, List, Union
import pandas as pd
from label_sleuth.data_access.core.data_structs import DisplayFields, LABEL_POSITIVE, Label, TextElement, LabelType, \
    URI_SEP

def get_element_group_by_texts(texts: Sequence[str], workspace_id, dataset_name, data_access, remove_duplicates) \
        -> Iterable[TextElement]:
    """
    The user may import a large number of labeled instances, and these will not necessarily be given with a text
    element uri. Thus, the goal here is to efficiently query for a group of text elements using only the texts, and
    optionally a *doc_uri* that the texts belong to.
    The order of the returned elements DOES NOT match the order of the texts given as input.
    """
    regex = '|'.join(f'^{re.escape(t)}$' for t in texts)
    elements = data_access.get_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                             sample_size=sys.maxsize, query=regex, is_regex=True,
                                             document_uri=None, remove_duplicates=remove_duplicates)
    return elements['results']


def merge_and_rename_dfs(a: pd.DataFrame, b: pd.DataFrame, on: Union[str, List[str]]):
    df = pd.merge(left=a, right=b, on=on, suffixes=('_x', None))
    df = df[df.columns.intersection(DisplayFields.__dict__)]
    return df

def process_labels_dataframe(workspace_id, dataset_name, data_access, labels_df_to_import: pd.DataFrame, apply_labels_to_duplicate_texts=False) \
        -> Dict[str, Dict[str, Dict[str, Label]]]:
    logging.warning("Currently label metadata is ignored")
    
    # replace punctuation with underscores in category names
    punctuation = string.punctuation.replace("'", "")

    if DisplayFields.doc_id in labels_df_to_import.columns:
        labels_df_to_import[DisplayFields.doc_id] = labels_df_to_import[DisplayFields.doc_id].apply(
            lambda x: x.replace(URI_SEP, '_'))
    labels_df_to_import[DisplayFields.category_name] = labels_df_to_import[DisplayFields.category_name].apply(str)
    labels_df_to_import[DisplayFields.category_name] = labels_df_to_import[DisplayFields.category_name].apply(
        lambda x: x.translate(x.maketrans(punctuation, '_' * len(punctuation))))
    # convert binary labels in dataframe to boolean
    positive_indicators = {LABEL_POSITIVE, str(LABEL_POSITIVE), str(LABEL_POSITIVE).upper(),
                           str(LABEL_POSITIVE).lower(), 1, "1"}
    labels_df_to_import[DisplayFields.label] = labels_df_to_import[DisplayFields.label].apply(
        lambda x: True if x in positive_indicators else False)
    if DisplayFields.label_type not in labels_df_to_import.columns:
        labels_df_to_import[DisplayFields.label_type] = LabelType.Standard.name

    # calling get_element_group_by_texts is the most expensive call on this function and,
    # removing duplicates from the texts iterable reduces significantly its runtime
    texts = labels_df_to_import[DisplayFields.text].unique()

    elements = get_element_group_by_texts(texts, workspace_id, dataset_name, data_access, remove_duplicates=False)   
        

    # convert Element list into a DataFrame in order to perform the join operation
    query_elements_df = pd.DataFrame(
            [[e.text, e.uri.split('-')[1], e.uri] for e in elements],
            columns=[DisplayFields.text, DisplayFields.doc_id, DisplayFields.uri]
        )


    def has_contradicting_labels(df: pd.DataFrame):
        return df[DisplayFields.label].unique().shape[0] > 1
  
    # check if the user uploaded contradicting labels
    group_by_cols = [DisplayFields.category_name, DisplayFields.text]
    if not apply_labels_to_duplicate_texts:
        group_by_cols.append(DisplayFields.doc_id)
    contradicting_labels_df = labels_df_to_import.groupby(group_by_cols).filter(has_contradicting_labels)
    contradicting_labels_texts = contradicting_labels_df.reset_index()[DisplayFields.text].unique().tolist()
    contracticting_labels_info = {
        'elements': contradicting_labels_texts
    }

    if apply_labels_to_duplicate_texts:
        # duplicates are dropped because if not two many duplicates will be present on the joined DataFrame
        labels_df_to_import.drop_duplicates(subset=[DisplayFields.category_name, DisplayFields.text], inplace=True)
        df = merge_and_rename_dfs(
                labels_df_to_import, 
                query_elements_df, 
                DisplayFields.text
            )
    else:
        df = merge_and_rename_dfs(
                labels_df_to_import, 
                query_elements_df, 
                [DisplayFields.doc_id, DisplayFields.text]
            )
    category_to_uri_to_label = {}
    for category_name, category_df in df.groupby(DisplayFields.category_name):
        uri_to_label = {}
        for uri, uri_df in category_df.groupby(DisplayFields.uri):
            for (label, label_type), _ in uri_df.groupby([DisplayFields.label, DisplayFields.label_type]):
                uri_to_label[uri] = {category_name: Label(label=bool(label), label_type=LabelType[label_type])}
        category_to_uri_to_label[category_name] = uri_to_label

    return category_to_uri_to_label, contracticting_labels_info



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from label_sleuth.models.core.catalog import ModelsCatalog
from label_sleuth.models.policy.static_model_policy import StaticModelPolicy


class ModelPolicies:
    """
    Model policies determine which type of classification model is used. Policies can be static, i.e. always return the
    same model type, or dynamic, e.g. a different model type is returned depending on the current iteration.
    """
    STATIC_SVM_ENSEMBLE = StaticModelPolicy(ModelsCatalog.SVM_ENSEMBLE)
    STATIC_SVM_WORD_EMBEDDINGS = StaticModelPolicy(ModelsCatalog.SVM_OVER_WORD_EMBEDDINGS)
    STATIC_SVM_GLOVE = StaticModelPolicy(ModelsCatalog.SVM_OVER_WORD_EMBEDDINGS)  # for backward compatibility
    STATIC_SVM_BOW = StaticModelPolicy(ModelsCatalog.SVM_OVER_BOW)
    STATIC_NB_WORD_EMBEDDINGS = StaticModelPolicy(ModelsCatalog.NB_OVER_WORD_EMBEDDINGS)
    STATIC_NB_BOW = StaticModelPolicy(ModelsCatalog.NB_OVER_BOW)
    STATIC_HF_BERT = StaticModelPolicy(ModelsCatalog.HF_BERT)
    STATIC_HF_XLM_ROBERTA = StaticModelPolicy(ModelsCatalog.HF_XLM_ROBERTA)




import time
import logging
import functools

def log_runtime(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        
        res = function(*args, **kwargs)

        end_time = time.time()
        total_time = round(end_time - start_time, 2)
        
        logging.info(f'{function.__name__} took {total_time} seconds, {round(total_time/60, 2)} minutes')

        return res

    return wrapper


import unittest
import tempfile
import os
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.models.util.disk_cache import save_model_prediction_store_to_disk, \
    load_model_prediction_store_from_disk


class TestDiskCache(unittest.TestCase):
    def test_save_equals_load(self):
        cache = {
            str(('SVM_mid', (('text', 'Embrace growth and innovation!'),)),): Prediction(True, 0.9),
            str(('SVM_mid', (('text', 'parking lot'),)),): Prediction(False, 0.4)
        }
        temp_dir = tempfile.TemporaryDirectory()
        cache_file_path = os.path.join(temp_dir.name, "cache.json")
        save_model_prediction_store_to_disk(cache_file_path, cache)
        loaded = load_model_prediction_store_from_disk(cache_file_path, Prediction)

        temp_dir.cleanup()
        self.assertEqual(cache, loaded)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
import random
import tempfile
import unittest
from datetime import datetime
from unittest.mock import patch

import pandas as pd

from label_sleuth.active_learning.core.active_learning_factory import ActiveLearningFactory
from label_sleuth.config import load_config
from label_sleuth.data_access.core.data_structs import DisplayFields, Document, Label, LABEL_NEGATIVE, LABEL_POSITIVE
from label_sleuth.data_access.file_based.file_based_data_access import FileBasedDataAccess
from label_sleuth.data_access.test_file_based_data_access import generate_corpus
from label_sleuth.models.core.model_api import ModelStatus
from label_sleuth.models.core.catalog import ModelsCatalog
from label_sleuth.models.core.models_factory import ModelFactory
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import OrchestratorStateApi, Iteration, \
    IterationStatus, ModelInfo
from label_sleuth.orchestrator.orchestrator_api import OrchestratorApi, NUMBER_OF_MODELS_TO_KEEP
from label_sleuth.training_set_selector.training_set_selector_factory import TrainingSetSelectionFactory


def add_random_labels_to_document(doc: Document, min_num_sentences_to_label: int, categories, seed=0):
    random.seed(seed)
    text_elements_to_label = random.sample(doc.text_elements, min(min_num_sentences_to_label, len(doc.text_elements)))
    for elem in text_elements_to_label:
        categories_to_label = random.sample(categories, random.randint(0, len(categories)))
        elem.category_to_label.update({cat: Label(label=random.sample([LABEL_POSITIVE, LABEL_NEGATIVE], 1)[0])
                                       for cat in categories_to_label})
    return text_elements_to_label


class TestOrchestratorAPI(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()
        background_jobs_manager = BackgroundJobsManager()
        cls.model_factory = ModelFactory(cls.temp_dir.name, background_jobs_manager, None)
        cls.active_learning_factory = ActiveLearningFactory()
        cls.data_access = FileBasedDataAccess(os.path.join(cls.temp_dir.name, "output"))
        cls.orchestrator_state = OrchestratorStateApi(os.path.join(cls.temp_dir.name, "output", "workspaces"))
        cls.training_set_selection_factory = TrainingSetSelectionFactory(cls.data_access, background_jobs_manager)
        cls.orchestrator_api = OrchestratorApi(cls.orchestrator_state, cls.data_access, cls.active_learning_factory,
                                               cls.model_factory, cls.training_set_selection_factory, background_jobs_manager,
                                               None,  # no need to use sentence embedding,
                                               load_config(os.path.abspath(os.path.join(__file__, os.pardir,
                                                                                        os.pardir,
                                                                                        "config_for_tests.json"))))

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()

    @patch.object(OrchestratorApi, 'set_labels')
    def test_import_category_labels(self, mock_set_labels):
        workspace_id = self.test_import_category_labels.__name__
        dataset_name = f'{workspace_id}_dump'
        categories = [f'{workspace_id}_cat_' + str(i) for i in range(8)]

        # create mock dataframe to be imported
        docs = generate_corpus(self.data_access, dataset_name, 5)
        self.orchestrator_api.create_workspace(workspace_id=workspace_id, dataset_name=dataset_name)
        elements_for_import = []
        for doc in docs:
            elements_for_import.extend(
                add_random_labels_to_document(doc, 1, categories))
        categories_with_labels = {cat for e in elements_for_import for cat in e.category_to_label.keys()}
        dicts_for_import = [{DisplayFields.uri: e.uri, DisplayFields.text: e.text, DisplayFields.category_name: cat,
                             DisplayFields.label: e.category_to_label[cat].label,
                             DisplayFields.label_type: e.category_to_label[cat].label_type.name}
                            for e in elements_for_import for cat in e.category_to_label.keys()
                            if len(e.category_to_label) > 0]
        df_for_import = pd.DataFrame(dicts_for_import)

        # create only one of the categories
        selected_cat1 = next(iter(categories_with_labels))
        category_id1 = self.orchestrator_api.create_new_category(workspace_id, selected_cat1, 'description')
        category_id_to_name = {category_id1: selected_cat1}

        def create_cat_id(ws, cat_name, desc):
            cat_id = len(category_id_to_name)+1
            category_id_to_name[cat_id] = cat_name
            return cat_id

        # call import_labels()
        with patch.object(OrchestratorApi, 'create_new_category') as mock_create_new_category:
            mock_create_new_category.side_effect = create_cat_id
            info_dict = self.orchestrator_api.import_category_labels(workspace_id, df_for_import)

        self.assertEqual(len(info_dict['categories']), len(categories_with_labels))

        # check the correct number of categories are being created
        self.assertEqual(mock_create_new_category.call_count, len(categories_with_labels) - 1)
        self.assertEqual(len(info_dict['categoriesCreated']), len(categories_with_labels) - 1)

        # check that the correct label info was sent to set_labels()
        label_dicts_sent = [set_labels_call.args[1] for set_labels_call in mock_set_labels.call_args_list]
        uris_imported = {d[DisplayFields.uri] for d in dicts_for_import}
        uris_sent = {uri for d in label_dicts_sent for uri in d.keys()}
        self.assertEqual(uris_imported, uris_sent)

        self.assertEqual(info_dict['total'], len(df_for_import))

        # check that all the (uri, category_name, label) combinations indeed came from the imported dataframe
        label_tuples_imported = \
            set(df_for_import[[DisplayFields.uri, DisplayFields.category_name, DisplayFields.label]].itertuples(
                name=None, index=False))
        label_tuples_sent = set()
        for d in label_dicts_sent:
            for uri, cat_to_label in d.items():
                for cat_id, label_obj in cat_to_label.items():
                    label_tuples_sent.add((uri, category_id_to_name[cat_id], label_obj.label))
        self.assertEqual(label_tuples_imported, label_tuples_sent)

    def test_export_and_import_workspace_labels(self):
        dataset_name = self.test_export_and_import_workspace_labels.__name__ + '_dump'
        doc = generate_corpus(self.data_access, dataset_name)[0]
        categories = ['cat_' + str(i) for i in range(3)]

        self.orchestrator_api.create_workspace(workspace_id='mock_workspace_1', dataset_name=dataset_name)
        category_ids = []
        for cat in categories:
            category_id = self.orchestrator_api.create_new_category('mock_workspace_1', cat, 'some_description')
            category_ids.append(category_id)
        labeled_elements_for_export = add_random_labels_to_document(doc, 5, category_ids)
        labeled_elements_for_export = [e for e in labeled_elements_for_export if e.category_to_label != {}]

        def mock_get_labeled_text_elements(wid, ds, category_id, *args, **kwargs):
            return {'results': [e for e in labeled_elements_for_export if category_id in e.category_to_label]}

        # use export_workspace_labels() to turn labeled_elements_for_export into a dataframe for export
        with patch.object(FileBasedDataAccess, 'get_labeled_text_elements', side_effect=mock_get_labeled_text_elements):
            exported_df = self.orchestrator_api.export_workspace_labels('mock_workspace_1', True)

        for column in exported_df.columns:
            self.assertIn(column, DisplayFields.__dict__.values())

        # import the resulting dataframe using import_category_labels()
        self.orchestrator_state.create_workspace(workspace_id='mock_workspace_2', dataset_name=dataset_name)
        self.orchestrator_api.import_category_labels('mock_workspace_2', exported_df)

        unique = set()
        labeled_elements_imported = \
            [element for cat_id in category_ids for element
             in self.orchestrator_api.get_all_labeled_text_elements('mock_workspace_2', dataset_name, cat_id)
             if element.uri not in unique and not unique.add(element.uri)]

        self.assertEqual(sorted(labeled_elements_for_export, key=lambda te: te.uri),
                         sorted(labeled_elements_imported, key=lambda te: te.uri))

    @patch.object(OrchestratorApi, 'run_iteration')
    @patch.object(OrchestratorStateApi, 'get_label_change_count_since_last_train')
    @patch.object(FileBasedDataAccess, 'get_label_counts')
    def test_train_if_recommended(self, mock_get_label_counts, mock_get_label_change_count, mock_run_iteration):
        workspace_id = self.test_train_if_recommended.__name__
        dataset_name = f'{workspace_id}_dump'
        category_name = f'{workspace_id}_cat'
        generate_corpus(self.data_access, dataset_name)
        self.orchestrator_api.create_workspace(workspace_id, dataset_name)
        category_id = self.orchestrator_api.create_new_category(workspace_id, category_name, 'some_description')

        change_threshold = self.orchestrator_api.config.changed_element_threshold
        first_model_pos_threshold = self.orchestrator_api.config.first_model_positive_threshold

        # do not trigger training
        negative_count = 1
        positive_count = max(max(0, change_threshold - negative_count - 1),
                             max(0, first_model_pos_threshold - 1))

        label_counts = {LABEL_POSITIVE: positive_count, LABEL_NEGATIVE: negative_count}
        mock_get_label_counts.return_value = label_counts
        mock_get_label_change_count.return_value = sum(label_counts.values())

        progress = self.orchestrator_api.get_progress(workspace_id, dataset_name, category_id)
        self.assertEqual(progress['all'], 100*positive_count/first_model_pos_threshold)

        self.orchestrator_api.train_if_recommended(workspace_id, category_id)
        mock_run_iteration.assert_not_called()

        # trigger training
        positive_count = max(change_threshold, first_model_pos_threshold) - negative_count + 1
        label_counts = {LABEL_POSITIVE: positive_count, LABEL_NEGATIVE: negative_count}
        mock_get_label_counts.return_value = label_counts
        mock_get_label_change_count.return_value = sum(label_counts.values())

        progress = self.orchestrator_api.get_progress(workspace_id, dataset_name, category_id)
        self.assertEqual(progress['all'], 100)

        train_set_selector = \
            self.orchestrator_api.training_set_selection_factory.get_training_set_selector(
                                      self.orchestrator_api.config.training_set_selection_strategy).__class__
        with patch.object(train_set_selector, 'get_train_set'):
            self.orchestrator_api.train_if_recommended(workspace_id, category_id)
        mock_run_iteration.assert_called()

    def test_set_label_increases_change_count(self):
        workspace_id = self.test_set_label_increases_change_count.__name__
        dataset_name = f'{workspace_id}_dump'
        category_name = f'{workspace_id}_cat'
        generate_corpus(self.data_access, dataset_name)
        self.orchestrator_api.create_workspace(workspace_id, dataset_name)
        category_id = self.orchestrator_api.create_new_category(workspace_id, category_name, 'some_description')
        text_elements = self.orchestrator_api.get_all_text_elements(dataset_name)
        self.orchestrator_api.set_labels(workspace_id, {text_elements[0].uri: {category_id: Label(LABEL_POSITIVE)}})
        num_changed = self.orchestrator_state.get_label_change_count_since_last_train(workspace_id, category_id)
        self.assertEqual(1, num_changed, msg="we set a label for one element")
        self.orchestrator_api.set_labels(workspace_id, {text_elements[0].uri: {category_id: Label(LABEL_POSITIVE)}})
        num_changed = self.orchestrator_state.get_label_change_count_since_last_train(workspace_id, category_id)
        self.assertEqual(2, num_changed, msg="we set a label for "
                                             "one element (same element should increase the change count)")

    def test_set_label_without_increasing_change_count(self):
        workspace_id = self.test_set_label_without_increasing_change_count.__name__
        dataset_name = f'{workspace_id}_dump'
        category_name = f'{workspace_id}_cat'
        generate_corpus(self.data_access, dataset_name)
        self.orchestrator_api.create_workspace(workspace_id, dataset_name)
        category_id = self.orchestrator_api.create_new_category(workspace_id, category_name, 'some_description')
        text_elements = self.orchestrator_api.get_all_text_elements(dataset_name)
        self.orchestrator_api.set_labels(workspace_id, {text_elements[0].uri: {category_id: Label(LABEL_POSITIVE)}},
                                         update_label_counter=False)
        num_changed = self.orchestrator_state.get_label_change_count_since_last_train(workspace_id, category_id)
        self.assertEqual(0, num_changed, msg="we set a label with update_label_counter=False "
                                             "so number of changed element should be zero")

    @patch.object(OrchestratorApi, 'delete_iteration_model')
    @patch.object(OrchestratorStateApi, 'get_all_iterations')
    def test_old_models_deletion(self, get_all_iterations, delete_iteration_model):
        workspace_id = self.test_old_models_deletion.__name__
        dataset_name = f'{workspace_id}_dump'
        category_name = f'{workspace_id}_cat'
        generate_corpus(self.data_access, dataset_name)
        self.orchestrator_api.create_workspace(workspace_id, dataset_name)
        category_id = self.orchestrator_api.create_new_category(workspace_id, category_name, 'some_description')
        get_all_iterations.return_value = \
            [Iteration(IterationStatus.READY, ModelInfo("x", ModelStatus.READY, datetime.now(), ModelsCatalog.RAND, {}),
                       {}, [])] * (NUMBER_OF_MODELS_TO_KEEP+1)
        self.orchestrator_api._delete_old_models(workspace_id, category_id, NUMBER_OF_MODELS_TO_KEEP-1)

        # _delete_old_models for iteration NUMBER_OF_MODELS_TO_KEEP-1 should not delete any model as there are
        # only NUMBER_OF_MODELS_TO_KEEP iterations in status READY
        delete_iteration_model.assert_not_called()

        self.orchestrator_api._delete_old_models(workspace_id, category_id, NUMBER_OF_MODELS_TO_KEEP)

        # _delete_old_models for iteration NUMBER_OF_MODELS_TO_KEEP should invoke delete_iteration_model for iteration 0
        delete_iteration_model.assert_called_with(workspace_id, category_id, 0)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import random
import tempfile
import unittest
from unittest.mock import MagicMock

from label_sleuth.models.core.languages import Languages
from label_sleuth.models.random_model import RandomModel
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

PREFIX = 'Fascinating sentence'


class TestModelAPI(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.model_api = RandomModel(self.temp_dir.name, BackgroundJobsManager())
        self.model_id, future = self.model_api.train([], Languages.ENGLISH, {})
        future.result()
        all_nums = list(range(1, 10))
        random.shuffle(all_nums)
        small_batch = random.sample(all_nums, 3)
        large_batch = all_nums
        self.sentences1 = [{'text': PREFIX + str(num)} for num in small_batch]
        self.sentences2 = [{'text': PREFIX + str(num)} for num in large_batch]

    def test_infer_no_cache(self):
        self.model_api.infer_by_id(model_id=self.model_id, items_to_infer=self.sentences1, use_cache=False)
        self.assertEqual(0, self.model_api.cache.get_current_size(), "cache size should be zero when use_cache=False")

    def test_additional_fields_are_part_of_cache_key(self):
        self.model_api.infer_by_id(model_id=self.model_id, items_to_infer=self.sentences1, use_cache=True)
        self.assertEqual(len(self.sentences1), self.model_api.cache.get_current_size(),
                         "cache size should be equals to the size of the inferred sentences")
        self.sentences1[0] = {**self.sentences1[0], "additional_field": "1"}
        self.model_api.infer_by_id(model_id=self.model_id, items_to_infer=self.sentences1, use_cache=True)
        self.assertEqual(len(self.sentences1)+1, self.model_api.cache.get_current_size(),
                         "now sentences[0] has an additional field so the cache size should be len(sentences)+1")

    def test_inferred_elements_cached(self):
        self.model_api.infer_by_id(model_id=self.model_id, items_to_infer=self.sentences1, use_cache=True)
        self.assertEqual(len(self.sentences1), self.model_api.cache.get_current_size(),
                         "cache size should be equal to the size of the inferred sentences")
        self.model_api._infer = MagicMock(name='_infer')

        self.model_api.infer_by_id(model_id=self.model_id, items_to_infer=self.sentences1, use_cache=True)
        self.model_api._infer.assert_not_called()

    def tearDown(self):
        self.temp_dir.cleanup()



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from dataclasses import dataclass, field
from enum import Enum

from typing import List, Tuple, Mapping

LABEL_POSITIVE = True
LABEL_NEGATIVE = False
BINARY_LABELS = frozenset({LABEL_NEGATIVE, LABEL_POSITIVE})
URI_SEP = "-"


class LabelType(Enum):
    Standard = 0
    Weak = 1


@dataclass
class Label:
    label: bool
    metadata: Mapping = field(default_factory=dict)
    label_type: LabelType = field(default=LabelType.Standard)

    def get_detailed_label_name(self):
        return str(self.label).lower() if self.label_type == LabelType.Standard \
            else f'{self.label_type.name}_{self.label}'.lower()

    def to_dict(self):
        dict_for_json = {'label': self.label, 'metadata': self.metadata, 'label_type': self.label_type.value}
        return dict_for_json

    def __post_init__(self):
        if type(self.label_type) == int:
            self.label_type = LabelType(self.label_type)


@dataclass
class TextElement:
    uri: str
    text: str
    span: List[Tuple]
    metadata: Mapping
    category_to_label: Mapping[int, Label]

    @classmethod
    def get_field_names(cls):
        return cls.__annotations__.keys()


@dataclass
class Document:
    uri: str
    text_elements: List[TextElement]
    metadata: Mapping


class DisplayFields:
    workspace_id = 'workspace_id'
    category_name = 'category_name'
    doc_id = 'document_id'
    dataset = 'dataset'
    text = 'text'
    uri = 'uri'
    element_metadata = 'element_metadata'
    label = 'label'
    label_metadata = 'label_metadata'  # TODO currently not supported
    label_type = 'label_type'
    csv_metadata_column_prefix = 'metadata_'



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import threading
import logging
from typing import Type

from label_sleuth.data_access.data_access_api import DataAccessApi
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager
from label_sleuth.training_set_selector.train_set_selector_api import TrainSetSelectorAPI


class TrainingSetSelectionFactory:
    """
    Given a model type, this factory returns the relevant implementation of ModelAPI
    """
    def __init__(self, data_access: DataAccessApi, background_jobs_manager: BackgroundJobsManager):
        self.loaded_strategies = {}
        self.data_access = data_access
        self.background_jobs_manager = background_jobs_manager
        self.lock = threading.RLock()

    def get_training_set_selector(self, strategy_type: Type[TrainSetSelectorAPI]) -> TrainSetSelectorAPI:
        with self.lock:
            if strategy_type not in self.loaded_strategies:
                try:
                    strategy_api = strategy_type(self.data_access, self.background_jobs_manager)
                    self.loaded_strategies[strategy_type] = strategy_api
                except Exception:
                    logging.exception(f"Could not get strategy type {strategy_type} from the training set selection"
                                      f" factory")

        return self.loaded_strategies[strategy_type]


#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from dataclasses import dataclass
from typing import Sequence


@dataclass
class Language:
    """
    This dataclass contains some parameters that are language-specific. The parameter values can be used by the
    classification model implementations, as well as by some general utility functions in the *analysis_utils* module.
    """
    name: str
    stop_words: Sequence[str]
    spacy_model_name: str = None
    fasttext_language_id: str = None
    right_to_left: bool = False

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __post_init__(self):
        if self.spacy_model_name is None and self.fasttext_language_id is None:
            raise Exception(f"Selected Language {self.name} should define either "
                            f"spacy_model_name or fasttext_language_id in class '{Language}'")


English = \
    Language(name='English',
             stop_words=["a", "about", "above", "after", "again", "against", "ain", "all", "am", "an", "and", "any",
                         "are", "aren", "aren't", "as", "at", "be", "because", "been", "before", "being", "below",
                         "between", "both", "but", "by", "can", "couldn", "couldn't", "d", "did", "didn", "didn't",
                         "do", "does", "doesn", "doesn't", "doing", "don", "don't", "down", "during", "each", "few",
                         "for", "from", "further", "had", "hadn", "hadn't", "has", "hasn", "hasn't", "have", "haven",
                         "haven't", "having", "he", "her", "here", "hers", "herself", "him", "himself", "his", "how",
                         "i", "if", "in", "into", "is", "isn", "isn't", "it", "it's", "its", "itself", "just", "ll",
                         "m", "ma", "me", "mightn", "mightn't", "more", "most", "mustn", "mustn't", "my", "myself",
                         "needn", "needn't", "no", "nor", "not", "now", "o", "of", "off", "on", "once", "only", "or",
                         "other", "our", "ours", "ourselves", "out", "over", "own", "re", "s", "same", "shan", "shan't",
                         "she", "she's", "should", "should've", "shouldn", "shouldn't", "so", "some", "such", "t",
                         "than", "that", "that'll", "the", "their", "theirs", "them", "themselves", "then", "there",
                         "these", "they", "this", "those", "through", "to", "too", "under", "until", "up", "ve", "very",
                         "was", "wasn", "wasn't", "we", "were", "weren", "weren't", "what", "when", "where", "which",
                         "while", "who", "whom", "why", "will", "with", "won", "won't", "wouldn", "wouldn't", "y",
                         "you", "you'd", "you'll", "you're", "you've", "your", "yours", "yourself", "yourselves",
                         "could", "he'd", "he'll", "he's", "here's", "how's", "i'd", "i'll", "i'm", "i've", "let's",
                         "ought", "she'd", "she'll", "that's", "there's", "they'd", "they'll", "they're", "they've",
                         "we'd", "we'll", "we're", "we've", "what's", "when's", "where's", "who's", "why's", "would"],
             spacy_model_name='en_core_web_lg')

Italiano = \
    Language(name='Italiano',
             stop_words=["scola", "varie", "saro", "stanno", "dagli", "certa", "le", "nemmeno", "molto", "avuti",
                         "posteriore", "facesse", "il", "nostra", "durante", "faceste", "dalla", "può", "della", "ora",
                         "faccia", "alla", "dice", "scorso", "probabilmente", "volta", "sareste", "tre", "fu", "dell",
                         "quanta", "soltanto", "fa", "degl", "diventare", "pure", "quantunque", "che", "anche", "pieno",
                         "nostri", "te", "facessero", "qualunque", "codesti", "nonsia", "come", "successivamente",
                         "stato", "mesi", "quelle", "insieme", "agli", "volte", "faresti", "nell", "lui", "ma",
                         "malgrado", "ossia", "allo", "lato", "mi", "aveste", "lo", "molti", "basta", "seguito",
                         "dallo", "cioe", "giacche", "avevate", "starebbero", "parecchio", "ansa", "paese", "partendo",
                         "secondo", "prima", "gliela", "quale", "là", "uno", "vari", "facciamo", "dove", "sarà",
                         "stati", "medesimo", "steste", "male", "stiate", "facciano", "ahime", "dovrà", "tra",
                         "stessero", "osi", "esse", "seguente", "stavo", "percio", "sarebbe", "avrai", "parecchie",
                         "codesto", "dall", "cominci", "tempo", "sui", "relativo", "faremo", "aver", "staremo", "citta",
                         "dei", "improvviso", "stesse", "verso", "generale", "po", "avreste", "mie", "parecchi",
                         "fuori", "uguali", "gruppo", "queste", "contro", "diventato", "col", "così", "puo", "sarete",
                         "quindi", "anni", "si", "stavamo", "mediante", "eravamo", "state", "stettero", "sto", "colui",
                         "facesti", "lungo", "altro", "fareste", "fummo", "agl", "solito", "certe", "abbia", "ai",
                         "faremmo", "stavano", "perfino", "no", "quelli", "certo", "miei", "feci", "fossi", "presa",
                         "purtroppo", "con", "facevi", "chi", "momento", "concernente", "stiamo", "ecc", "saremo",
                         "successivo", "una", "per", "coloro", "governo", "nessuno", "dappertutto", "conclusione",
                         "fra", "farà", "glieli", "sotto", "salvo", "dentro", "altri", "sarebbero", "fece", "tutto",
                         "quel", "effettivamente", "luogo", "alle", "essi", "ti", "ella", "sembrare", "casa", "foste",
                         "fatto", "sarò", "starò", "la", "avuta", "tu", "codesta", "in", "sugl", "staresti", "starà",
                         "fanno", "starete", "mila", "degli", "stesti", "avute", "voi", "esser", "finalmente", "avemmo",
                         "ebbe", "peggio", "cosa", "nondimeno", "maggior", "lei", "grande", "bene", "avevo", "ore",
                         "stava", "nello", "stiano", "allora", "farò", "infatti", "sembrato", "dirimpetto", "nella",
                         "me", "ero", "piglia", "altrove", "nel", "saremmo", "ho", "entrambi", "ognuno", "altrimenti",
                         "tanto", "avranno", "dal", "recentemente", "tali", "sulle", "tua", "dovunque", "farebbe",
                         "sei", "avessimo", "pero", "avremmo", "coll", "facevate", "spesso", "avevano", "cio", "niente",
                         "inoltre", "ancora", "otto", "li", "abbastanza", "varia", "qualcuno", "vale", "già", "ieri",
                         "vostra", "cento", "facemmo", "stette", "faceva", "persone", "avevamo", "ci", "troppo", "però",
                         "dunque", "sulla", "avrete", "all", "vi", "piuttosto", "saresti", "stessi", "ottanta", "bravo",
                         "sia", "ed", "proprio", "intanto", "quello", "nelle", "chiunque", "facevano", "visto", "colei",
                         "miliardi", "avrei", "press", "invece", "nulla", "ahimè", "altrui", "se", "sono", "sua",
                         "gliele", "consiglio", "ulteriore", "peccato", "recente", "tuoi", "lavoro", "frattempo",
                         "avrà", "qualcosa", "eppure", "sopra", "stando", "forza", "fossero", "attraverso",
                         "nonostante", "gliene", "dai", "negli", "farete", "moltissimo", "poiche", "sempre", "quante",
                         "noi", "sembri", "dietro", "marche", "poco", "loro", "nei", "vostre", "giorno", "ad", "città",
                         "deve", "novanta", "oltre", "eri", "grazie", "milioni", "quest", "talvolta", "cui", "glielo",
                         "dalle", "lasciato", "certi", "facessi", "solo", "avrebbero", "essendo", "quanti", "stia",
                         "hai", "stemmo", "via", "dov", "staremmo", "avente", "nove", "perciò", "quella", "mai", "tuo",
                         "nostre", "lontano", "ex", "nostro", "ciascuno", "fino", "perché", "ultimo", "faranno", "ha",
                         "inc", "perche", "al", "ognuna", "sara", "modo", "ciascuna", "faccio", "qui", "chicchessia",
                         "cos", "mentre", "adesso", "sig", "hanno", "cosi", "avrò", "persino", "dovra", "avresti",
                         "facendo", "favore", "benissimo", "stareste", "sul", "da", "stata", "co", "quanto", "tale",
                         "facessimo", "avanti", "dagl", "tuttavia", "farai", "quando", "avendo", "preferibilmente",
                         "avremo", "starebbe", "finche", "caso", "stetti", "malissimo", "essere", "giorni", "stesso",
                         "vostri", "vario", "trenta", "ecco", "fosti", "stai", "ne", "mia", "furono", "vita", "sarei",
                         "poi", "anticipo", "delle", "tranne", "principalmente", "coi", "stavate", "potrebbe", "farei",
                         "siete", "brava", "fin", "cima", "attesa", "stavi", "del", "fecero", "avesse", "forse",
                         "detto", "gia", "farebbero", "avessero", "saranno", "minimi", "piu", "senza", "starei",
                         "sarai", "possedere", "non", "pochissimo", "un", "primo", "piedi", "davanti", "titolo", "od",
                         "uomo", "stessimo", "mio", "negl", "tutte", "avrebbe", "suo", "fossimo", "fai", "io", "mosto",
                         "parte", "facevo", "questa", "facevamo", "tutta", "sette", "gli", "fui", "scopo", "vostro",
                         "fine", "fosse", "trovato", "nessuna", "qualcuna", "riecco", "cogli", "sta", "possa", "ebbero",
                         "su", "avete", "fare", "nuovo", "posto", "a", "realmente", "meno", "registrazione", "quasi",
                         "mondo", "sullo", "comunque", "siano", "dopo", "esempio", "avevi", "siamo", "avesti", "nessun",
                         "tue", "va", "cortesia", "intorno", "questi", "ministro", "anno", "alcuni", "sugli", "tutti",
                         "srl", "diventa", "subito", "nazionale", "siate", "ebbi", "sue", "alcuna", "abbiate", "mezzo",
                         "due", "meglio", "eravate", "ogni", "macche", "dello", "magari", "più", "alcuno", "questo",
                         "avere", "oggi", "vicino", "futuro", "circa", "staranno", "stessa", "di", "avuto", "haha",
                         "assai", "starai", "abbiamo", "era", "oppure", "quali", "accidenti", "abbiano", "suoi", "egli",
                         "neppure", "averlo", "qualche", "sull", "avessi", "quattro", "facciate", "dire", "mancanza",
                         "affinche", "erano", "conciliarsi", "sembra", "aveva"],
             spacy_model_name='it_core_news_lg')

Romanian = \
    Language(name='Romanian',
             stop_words=["a", "abia", "acea", "aceasta", "această", "aceea", "aceeasi", "acei", "aceia", "acel",
                         "acela", "acelasi", "acele", "acelea", "acest", "acesta", "aceste", "acestea", "acestei",
                         "acestia", "acestui", "aceşti", "aceştia", "acolo", "acord", "acum", "adica", "ai", "aia",
                         "aibă", "aici", "aiurea", "al", "ala", "alaturi", "ale", "alea", "alt", "alta", "altceva",
                         "altcineva", "alte", "altfel", "alti", "altii", "altul", "am", "anume", "apoi", "ar", "are",
                         "as", "asa", "asemenea", "asta", "astazi", "astea", "astfel", "astăzi", "asupra", "atare",
                         "atat", "atata", "atatea", "atatia", "ati", "atit", "atita", "atitea", "atitia", "atunci",
                         "au", "avea", "avem", "aveţi", "avut", "azi", "aş", "aşadar", "aţi", "b", "ba", "bine",
                         "bucur", "bună", "c", "ca", "cam", "cand", "capat", "care", "careia", "carora", "caruia",
                         "cat", "catre", "caut", "ce", "cea", "ceea", "cei", "ceilalti", "cel", "cele", "celor", "ceva",
                         "chiar", "ci", "cinci", "cind", "cine", "cineva", "cit", "cita", "cite", "citeva", "citi",
                         "citiva", "conform", "contra", "cu", "cui", "cum", "cumva", "curând", "curînd", "când", "cât",
                         "câte", "câtva", "câţi", "cînd", "cît", "cîte", "cîtva", "cîţi", "că", "căci", "cărei",
                         "căror", "cărui", "către", "d", "da", "daca", "dacă", "dar", "dat", "datorită", "dată", "dau",
                         "de", "deasupra", "deci", "decit", "degraba", "deja", "deoarece", "departe", "desi", "despre",
                         "deşi", "din", "dinaintea", "dintr", "dintr-", "dintre", "doar", "doi", "doilea", "două",
                         "drept", "dupa", "după", "dă", "e", "ea", "ei", "el", "ele", "era", "eram", "este", "eu",
                         "exact", "eşti", "f", "face", "fara", "fata", "fel", "fi", "fie", "fiecare", "fii", "fim",
                         "fiu", "fiţi", "foarte", "fost", "frumos", "fără", "g", "geaba", "graţie", "h", "halbă", "i",
                         "ia", "iar", "ieri", "ii", "il", "imi", "in", "inainte", "inapoi", "inca", "incit", "insa",
                         "intr", "intre", "isi", "iti", "j", "k", "l", "la", "le", "li", "lor", "lui", "lângă", "lîngă",
                         "m", "ma", "mai", "mare", "mea", "mei", "mele", "mereu", "meu", "mi", "mie", "mine", "mod",
                         "mult", "multa", "multe", "multi", "multă", "mulţi", "mulţumesc", "mâine", "mîine", "mă", "n",
                         "ne", "nevoie", "ni", "nici", "niciodata", "nicăieri", "nimeni", "nimeri", "nimic", "niste",
                         "nişte", "noastre", "noastră", "noi", "noroc", "nostri", "nostru", "nou", "noua", "nouă",
                         "noştri", "nu", "numai", "o", "opt", "or", "ori", "oricare", "orice", "oricine", "oricum",
                         "oricând", "oricât", "oricînd", "oricît", "oriunde", "p", "pai", "parca", "patra", "patru",
                         "patrulea", "pe", "pentru", "peste", "pic", "pina", "plus", "poate", "pot", "prea", "prima",
                         "primul", "prin", "printr-", "putini", "puţin", "puţina", "puţină", "până", "pînă", "r", "rog",
                         "s", "sa", "sa-mi", "sa-ti", "sai", "sale", "sau", "se", "si", "sint", "sintem", "spate",
                         "spre", "sub", "sunt", "suntem", "sunteţi", "sus", "sută", "sînt", "sîntem", "sînteţi", "să",
                         "săi", "său", "t", "ta", "tale", "te", "ti", "timp", "tine", "toata", "toate", "toată",
                         "tocmai", "tot", "toti", "totul", "totusi", "totuşi", "toţi", "trei", "treia", "treilea", "tu",
                         "tuturor", "tăi", "tău", "u", "ul", "ului", "un", "una", "unde", "undeva", "unei", "uneia",
                         "unele", "uneori", "unii", "unor", "unora", "unu", "unui", "unuia", "unul", "v", "va", "vi",
                         "voastre", "voastră", "voi", "vom", "vor", "vostru", "vouă", "voştri", "vreme", "vreo",
                         "vreun", "vă", "x", "z", "zece", "zero", "zi", "zice", "îi", "îl", "îmi", "împotriva", "în",
                         "înainte", "înaintea", "încotro", "încât", "încît", "între", "întrucât", "întrucît", "îţi",
                         "ăla", "ălea", "ăsta", "ăstea", "ăştia", "şapte", "şase", "şi", "ştiu", "ţi", "ţie"],
             spacy_model_name='ro_core_news_lg')

Hebrew = Language(name='Hebrew',
                  stop_words=["זה", "זאת", "אלה", "אלו", "גם", "של", "מה", "או"],
                  fasttext_language_id="he",
                  right_to_left=True)


Arabic = Language(name='Arabic',
                  stop_words=["التى", "التي", "الذى", "الذي", "الذين", "ذلك", "هذا", "هذه", "هؤلاء", "قد", "وقد", "حيث",
                              "ان", "إن", "انه", "وان", "فان", "فإن", "بان", "اي", "أي", "ايضا", "أيضا", "إياه"],
                  fasttext_language_id='ar',
                  right_to_left=True)


class Languages:
    ENGLISH = English
    ITALIANO = Italiano
    ROMANIAN = Romanian
    HEBREW = Hebrew
    ARABIC = Arabic

    @classmethod
    def all_languages(cls):
        return {value for name, value in vars(cls).items() if name.isupper()}



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc
import sys
from dataclasses import dataclass
from enum import Enum

from typing import Iterable, Sequence, Mapping, List, Union, Set
from label_sleuth.data_access.core.data_structs import Document, TextElement, Label, URI_SEP, LabelType


class AlreadyExistsException(Exception):
    def __init__(self, message, documents):
        self.message = message
        self.documents = documents


@dataclass
class DocumentStatistics:
    documents_loaded: int
    text_elements_loaded: int


class LabeledStatus(Enum):
    UNLABELED = 0
    LABELED = 1
    ALL = 2


def get_document_uri(uri):
    """
    get document uri from element uri
    """
    return uri.rsplit(URI_SEP, 1)[0]


class DataAccessApi(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def add_documents(self, dataset_name: str, documents: Iterable[Document]) -> DocumentStatistics:
        """
        Add new documents to a given dataset; If dataset does not exist, create it.

        :param dataset_name: the name of the dataset to which the documents should be added.
        :param documents: an Iterable over Document type.
        """

    @abc.abstractmethod
    def set_labels(self, workspace_id: str, uris_to_labels: Mapping[str, Mapping[int, Label]],
                   apply_to_duplicate_texts=False):
        """
        Set labels to TextElements in dataset for a given workspace_id.

        :param workspace_id: the workspace_id of the labeling effort.
        :param uris_to_labels: list of tuples of TextElement URI and a dict that represents a label.
        The dict keys are category ids and values are Labels. For example: [(uri_1, {0: Label_cat_1}),
                                                                            (uri_2, {0: Label_cat_1,
                                                                                     1: Label_cat_2})]
        :param apply_to_duplicate_texts: if True, also set the same labels for additional URIs that are duplicates
        of the URIs provided.
        """

    @abc.abstractmethod
    def unset_labels(self, workspace_id: str, category_id: int, uris: Sequence[str], apply_to_duplicate_texts=False):
        """
        Remove workspace labels for a certain category from a specified list of uris.

        :param workspace_id: the workspace_id of the labeling effort.
        :param category_id: the id of the category labels are assigned to.
        :param uris: list of URIs to unset the label for.
        :param apply_to_duplicate_texts: if True, also unset the same labels for additional URIs that have the same text
        as that of the URIs provided.
        """

    @abc.abstractmethod
    def get_documents(self, workspace_id: Union[None, str], dataset_name: str, uris: Iterable[str],
                      label_types: Union[None,Set[LabelType]] = frozenset({LabelType.Standard})) \
            -> List[Document]:
        """
        Return a List of Documents, from the given dataset_name, matching the uris provided, and add the label
        information for the workspace to the TextElements of these Documents, if available.

        :param workspace_id: the workspace_id of the labeling effort. if None, documents are returned without
        label information
        :param dataset_name: the name of the dataset from which the documents should be retrieved.
        :param uris: an Iterable of uris of Documents, represented as string.
        :param label_types:  If workspace_id is provided, select which label_types to retrieve
                             by default, only the LabelType.Standard (strong labels) are retrieved.
        :return: a List of Document objects, from the given dataset_name, matching the uris provided, containing label
        information for the TextElements of these Documents, if available.
        """

    @abc.abstractmethod
    def get_all_document_uris(self, dataset_name: str) -> List[str]:
        """
        Return a List of all Document uris in the given dataset_name.

        :param dataset_name: the name of the dataset from which the Document uris should be retrieved.
        :return: a List of all Document uris in the given dataset_name.
        """

    @abc.abstractmethod
    def get_all_text_elements_uris(self, dataset_name: str) -> List[str]:
        """
        Return a List of all TextElement uris in the given dataset_name.

        :param dataset_name: the name of the dataset from which the TextElement uris should be retrieved.
        :return: a List of all TextElement uris in the given dataset_name.
        """

    @abc.abstractmethod
    def get_text_element_count(self, dataset_name: str) -> int:
        """
        Return the total number of TextElements in the given dataset_name.
        :param dataset_name:
        """

    @abc.abstractmethod
    def get_all_text_elements(self, dataset_name: str) -> List[TextElement]:
        """
        Return a List of all TextElement in the given dataset_name.

        :param dataset_name: the name of the dataset from which the TextElement should be retrieved.
        """

    @abc.abstractmethod
    def get_text_elements(self, workspace_id: str, dataset_name: str, sample_size: int = sys.maxsize,
                          sample_start_idx: int = 0, query: str = None, is_regex: bool = False,
                          remove_duplicates=False, document_uri=None, random_state: int = 0) -> Mapping:
        """
        Sample *sample_size* TextElements from dataset_name, optionally limiting to those matching a query,
        and add their labels information for workspace_id, if available.

        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which TextElements are sampled
        :param sample_size: how many TextElements should be sampled
        :param sample_start_idx: get elements starting from this index (for pagination). Default is 0
        :param query: a query string to search for in the sampled TextElements.
                      If None, then no such filtering is performed.
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param document_uri: get elements from a particular document
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """

    @abc.abstractmethod
    def get_unlabeled_text_elements(self, workspace_id: str, dataset_name: str, category_id: int,
                                    sample_size: int = sys.maxsize, sample_start_idx: int = 0,
                                    query: str = None, is_regex: bool = False,
                                    remove_duplicates=False, random_state: int = 0) -> Mapping:
        """
        Sample *sample_size* TextElements from dataset_name, unlabeled for category_id in workspace_id, optionally
        limiting to those matching a query.

        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which TextElements are sampled
        :param category_id: we demand that the elements are not labeled for this category
        :param sample_size: how many TextElements should be sampled
        :param sample_start_idx: get elements starting from this index (for pagination). Default is 0
        :param query: a query string to search for in the sampled TextElements.
                      If None, then no such filtering is performed.
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """

    @abc.abstractmethod
    def get_labeled_text_elements(self, workspace_id: str, dataset_name: str, category_id: int,
                                  sample_size: int = sys.maxsize,
                                  query: str = None, is_regex: bool = False,
                                  remove_duplicates=False, random_state: int = 0,
                                  label_types: Set[LabelType] = frozenset({LabelType.Standard})) -> Mapping:
        """
        Sample *sample_size* TextElements from dataset_name, labeled for category_id in workspace_id,
        optionally limiting to those matching a query.

        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which TextElements are sampled
        :param category_id: we demand that the elements are labeled for this category
        :param sample_size: how many TextElements should be sampled
        :param query: a query string to search for in the sampled TextElements.
                      If None, then no such filtering is performed.
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :param label_types: by default, only the LabelType.Standard (strong labels) are retrieved.
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """

    @abc.abstractmethod
    def get_label_counts(self, workspace_id: str, dataset_name: str, category_id: int, remove_duplicates=False,
                         label_types: Set[LabelType] = frozenset(LabelType._member_map_.values()),
                         fine_grained_counts=True) -> Mapping[Union[str, bool], int]:
        """
        Return for each label value, assigned to category_id, the total count of its appearances in dataset_name.
        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which labels count should be generated
        :param category_id: the id of the category whose label information is the target
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param label_types: by default, labels of all types are retrieved.
        :param fine_grained_counts: if True, count labels of each label type separately.
        :return: a map whose keys are label values, and the values are the number of TextElements this label was
        assigned to.
        """

    @abc.abstractmethod
    def delete_all_labels(self, workspace_id, dataset_name):
        """
        Delete the labels info of the given workspace_id for the given dataset (other labels info files are kept).
        :param workspace_id:
        :param dataset_name:
        """

    @abc.abstractmethod
    def delete_labels_for_category(self, workspace_id: str, dataset_name: str, category_id: int):
        """
        Delete the labels info associated with the given category.
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        """

    @abc.abstractmethod
    def get_text_elements_by_uris(self, workspace_id: str, dataset_name: str, uris: Iterable[str],
                                  label_types: Set[LabelType] = frozenset({LabelType.Standard})) -> List[TextElement]:
        """
        Return a List of TextElement objects from the given dataset_name, matching the uris provided, and add the label
        information for the workspace to these TextElements, if available.
        :param workspace_id:
        :param dataset_name:
        :param uris:
        :param label_types:  by default, only the LabelType.Standard (strong labels) are retrieved.
        """

    @abc.abstractmethod
    def get_text_element_iterator(self, workspace_id, dataset_name, shuffle=False, random_state: int = 0,
                                  remove_duplicates=False) -> Iterable[TextElement]:
        """
        Enables iterating over TextElement objects from the given *dataset_name*. Particularly useful where the total
        number of elements required is not known in advance.
        :param workspace_id:
        :param dataset_name:
        :param shuffle: if True, the iterator yields the text elements in random order.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        """

    @abc.abstractmethod
    def get_all_dataset_names(self) -> List[str]:
        """
        :return: a list of all available datset names
        """

    @abc.abstractmethod
    def delete_dataset(self, dataset_name: str):
        """
        Delete dataset by name
        :param dataset_name:
        """

    @abc.abstractmethod
    def preload_dataset(self, dataset_name):
        """
        Preload dataset to memory
        :param dataset_name:
        """



from label_sleuth.data_access.core.data_structs import LabelType
from label_sleuth.training_set_selector.train_set_selectors import TrainSetSelectorAllLabeled, \
    TrainSetSelectorEnforcePositiveNegativeRatio


class TrainSetSelectorAllLabeledIncludeWeak(TrainSetSelectorAllLabeled):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard, LabelType.Weak})

class TrainSetSelectorEqualRatio(TrainSetSelectorEnforcePositiveNegativeRatio):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard},
                                                            required_negative_ratio=1)

class TrainSetSelectorX2NegativeRatio(TrainSetSelectorEnforcePositiveNegativeRatio):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard},
                                                            required_negative_ratio=2)

class TrainSetSelectorX10NegativeRatio(TrainSetSelectorEnforcePositiveNegativeRatio):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard},
                                                            required_negative_ratio=10)


class TrainSetSelectorEqualRatioIncludeWeak(TrainSetSelectorEnforcePositiveNegativeRatio):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard, LabelType.Weak},
                         required_negative_ratio=1)


class TrainSetSelectorX2NegativeRatioIncludeWeak(TrainSetSelectorEnforcePositiveNegativeRatio):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard, LabelType.Weak},
                         required_negative_ratio=2)


class TrainSetSelectorX10NegativeRatioIncludeWeak(TrainSetSelectorEnforcePositiveNegativeRatio):
    def __init__(self, data_access, background_jobs_manager):
        super().__init__(data_access, background_jobs_manager, label_types={LabelType.Standard, LabelType.Weak},
                         required_negative_ratio=10)


class TrainSetSelectorsCatalog:
    """
    Given the current set of elements labeled by the user, a TrainingSetSelectionStrategy determines which examples
    will be sent in practice to the model at training time. For example, the strategy may specify that additional
    _unlabeled_ elements will be given to the model as weak negative examples.
    """
    ALL_LABELED = TrainSetSelectorAllLabeled
    ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_EQUAL_RATIO = TrainSetSelectorEqualRatio
    ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_X2_RATIO = TrainSetSelectorX2NegativeRatio
    ALL_LABELED_PLUS_UNLABELED_AS_NEGATIVE_X10_RATIO = TrainSetSelectorX10NegativeRatio
    ALL_LABELED_INCLUDE_WEAK = TrainSetSelectorAllLabeledIncludeWeak
    ALL_LABELED_INCLUDE_WEAK_PLUS_UNLABELED_AS_NEGATIVE_EQUAL_RATIO = TrainSetSelectorEqualRatioIncludeWeak
    ALL_LABELED_INCLUDE_WEAK_PLUS_UNLABELED_AS_NEGATIVE_X2_RATIO = TrainSetSelectorX2NegativeRatioIncludeWeak
    ALL_LABELED_INCLUDE_WEAK_PLUS_UNLABELED_AS_NEGATIVE_X10_RATIO = TrainSetSelectorX10NegativeRatioIncludeWeak



import os
import tempfile

import jsonpickle

from label_sleuth.models.core.model_api import ModelAPI
from label_sleuth.models.core.models_factory import ModelFactory
from label_sleuth.models.core.tools import SentenceEmbeddingService
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager


def get_model_api(model_path: str, sentence_embedding_model_path=os.getcwd()) -> ModelAPI:
    """
    Get an instance of ModelAPI according to the provided model_type
    For external use only, do not use inside sleuth project.

    :param model_path: directory of exported model, used for reading the model type
    :param sentence_embedding_model_path: Where to save sentence embedding model if
                                          used by the model type. Defaults to os.getcwd().
    """
    model_info_path = os.path.join(model_path, "model_info.json")
    with open(model_info_path) as json_file:
        model_info = json_file.read()
    model_info = jsonpickle.decode(model_info)
    model_factory = ModelFactory(output_dir=tempfile.gettempdir(),
                                 background_jobs_manager=BackgroundJobsManager(),
                                 sentence_embedding_service=SentenceEmbeddingService(
                                     embedding_model_dir=sentence_embedding_model_path))
    return model_factory.get_model_api(model_info["model_type"])




#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import List

import numpy as np

from label_sleuth.active_learning.core.active_learning_api import ActiveLearningStrategy
from label_sleuth.active_learning.policy.active_learning_policy import ActiveLearningPolicy


class ActiveLearningChangingPolicy(ActiveLearningPolicy):
    """
    A dynamic active learning policy that follows a predefined pattern of switching between active learning strategies
    after a certain number of iterations.
    """

    def __init__(self, active_learning_strategies: List[ActiveLearningStrategy],
                 num_iterations_per_strategy: List[int]):
        """
        :param active_learning_strategies: a list of the N strategies to be used by the policy
        :param num_iterations_per_strategy: a corresponding list specifying the number of iterations for each of the
        first N-1 strategies to be used; the Nth strategy type will be used for all subsequent iterations.
        For example, if there are 3 strategies, and num_iterations_per_strategy=[5, 2], then strategy A will be used for
        iterations 0-4, strategy B for iterations 5-6, and strategy C from iteration 7 onwards.
        """
        if len(active_learning_strategies) != len(num_iterations_per_strategy) + 1:
            raise Exception(
                f"The number of strategies provided ({len(active_learning_strategies)}) does not match the provided "
                f"list of {len(num_iterations_per_strategy)} active learning strategy switch points. "
                f"For each active learning type, except the last one, the number of iterations for this active "
                f"learning strategy to be used must be specified.")
        self.active_learning_strategies = active_learning_strategies
        self.num_iterations_per_strategy = num_iterations_per_strategy
        self.switch_points = np.cumsum(num_iterations_per_strategy)

    def get_active_learning_strategy(self, iteration_num: int) -> ActiveLearningStrategy:
        for active_learning_strategy, switch_point in zip(self.active_learning_strategies, self.switch_points):
            if iteration_num < switch_point:
                return active_learning_strategy
        return self.active_learning_strategies[-1]

    def get_name(self):
        name = ""
        for active_learning_strategy, n_iter in zip(self.active_learning_strategies, self.num_iterations_per_strategy):
            name += f'{active_learning_strategy.name}x{n_iter}-'
        name += f"{self.active_learning_strategies[-1].name}"
        return name



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import Sequence

from label_sleuth.active_learning.core.active_learning_api import ActiveLearner
from label_sleuth.data_access.core.data_structs import TextElement
from label_sleuth.models.core.prediction import Prediction


class HybridLearner(ActiveLearner):
    """
    Using this module, it possible to combine different active learning strategies. Two source AL modules are used to
    initialize the HybridLearner, and active learning scores for each example are calculated by averaging the
    active learning scores given by the two source modules.
    """
    def __init__(self, active_learner1, active_learner2):
        self.active_learner1 = active_learner1
        self.active_learner2 = active_learner2

    def get_per_element_score(self, candidate_text_elements: Sequence[TextElement],
                              candidate_text_element_predictions: Sequence[Prediction], workspace_id: str,
                              dataset_name: str, category_id: int) -> Sequence[float]:
        scores1 = self.active_learner1.get_per_element_score(candidate_text_elements,
                                                             candidate_text_element_predictions,
                                                             workspace_id, dataset_name, category_id)
        scores2 = self.active_learner2.get_per_element_score(candidate_text_elements,
                                                             candidate_text_element_predictions,
                                                             workspace_id, dataset_name, category_id)
        scores = [(s1 + s2) / 2 for s1, s2 in zip(scores1, scores2)]

        return scores



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc

from concurrent.futures import Future
from typing import Sequence, Set

from label_sleuth.data_access.core.data_structs import TextElement, LabelType
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager



class TrainSetSelectorAPI(object, metaclass=abc.ABCMeta):

    def __init__(self, data_access, background_jobs_manager: BackgroundJobsManager, gpu_support=False,
                 label_types=frozenset({LabelType.Standard})):
        self.data_access = data_access
        self.label_types = label_types
        self.background_jobs_manager = background_jobs_manager
        self.gpu_support = gpu_support

    def collect_train_set(self, workspace_id: str, train_dataset_name: str, category_id: int, category_name: str,
                          category_description:str, done_callback=None) \
            -> Future:
        future = self.background_jobs_manager.add_background_job(
            self.get_train_set, args=(workspace_id, train_dataset_name, category_id, category_name,
                                      category_description),
            use_gpu=self.gpu_support, done_callback=done_callback)
        return future

    @abc.abstractmethod
    def get_train_set(self, workspace_id: str, train_dataset_name: str, category_id: int,
                      category_name: str, category_description: str) -> Sequence[TextElement]:
        """
        For a given workspace, dataset and category, prepare and return a train set for training the model.
        Returns a list of TextElement objects (containing labels for the category, and possibly metadata about
        some of these labels).

        :param workspace_id:
        :param train_dataset_name:
        :param category_id:
        :param category_name:
        :param category_description:
        """

    def get_label_types(self) -> Set[LabelType]:
        """
        Return the types of labels this selector is using
        """
        return set(self.label_types)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import os
import pickle

from dataclasses import dataclass
from typing import Union

import numpy as np

import sklearn.svm
from sklearn.feature_extraction.text import CountVectorizer

from label_sleuth.models.core.languages import Language, Languages
from label_sleuth.models.core.model_api import ModelAPI
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.models.core.tools import RepresentationType, SentenceEmbeddingService
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s')

@dataclass
class SVMModelComponents:
    model: Union[sklearn.svm.LinearSVC, sklearn.svm.SVC]
    vectorizer: None
    language: Language


class SVM(ModelAPI):
    def __init__(self, output_dir, representation_type: RepresentationType,
                 background_jobs_manager: BackgroundJobsManager, sentence_embedding_service: SentenceEmbeddingService,
                 kernel="linear"):
        super().__init__(output_dir, background_jobs_manager)
        self.kernel = kernel
        self.representation_type = representation_type
        if self.representation_type == RepresentationType.WORD_EMBEDDING:
            self.sentence_embedding_service = sentence_embedding_service

    def _train(self, model_id, train_data, model_params):
        if self.kernel == "linear":
            model = sklearn.svm.LinearSVC()
        elif self.kernel == "rbf":
            logging.warning("Using some arbitrary low gamma, gamma and C values might be better with tuning")
            model = sklearn.svm.SVC(gamma=1e-08)
        else:
            raise ValueError("Unknown kernel type")

        language = self.get_language(self.get_model_dir_by_id(model_id))
        texts = [x['text'] for x in train_data]
        train_data_features, vectorizer = self.input_to_features(texts, language=language)
        labels = np.array([x['label'] for x in train_data])

        model.fit(train_data_features, labels)

        with open(os.path.join(self.get_model_dir_by_id(model_id), "vectorizer"), "wb") as fl:
            pickle.dump(vectorizer, fl)
        with open(os.path.join(self.get_model_dir_by_id(model_id), "model"), "wb") as fl:
            pickle.dump(model, fl)

    def load_model(self, model_path) -> SVMModelComponents:
        with open(os.path.join(model_path, "model"), "rb") as fl:
            model = pickle.load(fl)
        with open(os.path.join(model_path, "vectorizer"), "rb") as fl:
            vectorizer = pickle.load(fl)
        language = self.get_language(model_path)

        return SVMModelComponents(model=model, vectorizer=vectorizer, language=language)

    def infer(self, model_components: SVMModelComponents, items_to_infer):
        features_all_texts, _ = self.input_to_features([x['text'] for x in items_to_infer],
                                                       language=model_components.language,
                                                       vectorizer=model_components.vectorizer)
        labels = model_components.model.predict(features_all_texts).tolist()
        # The True label is in the second position as sorted([True, False]) is [False, True]
        scores = [probs[1] for probs in self.get_probs(model_components.model, features_all_texts)]
        return [Prediction(label=label, score=score) for label, score in zip(labels, scores)]

    @staticmethod
    def get_probs(model, features):
        distances = np.array(model.decision_function(features))  # get distances from hyperplanes (per class)
        if len(distances.shape) == 1:  # binary classification
            distances = distances / 2 + 0.5
            distances = np.expand_dims(distances, 1)
            distances = np.concatenate([1 - distances, distances], axis=1)
        # softmax to convert distances to probabilities
        prob = np.exp(distances) / np.sum(np.exp(distances), axis=1, keepdims=True)
        return prob

    def input_to_features(self, texts, language=Languages.ENGLISH, vectorizer=None):
        if self.representation_type == RepresentationType.BOW:
            if vectorizer is None:
                vectorizer = CountVectorizer(analyzer="word", tokenizer=None, preprocessor=None, stop_words=None,
                                             lowercase=True, max_features=10000)
                train_data_features = vectorizer.fit_transform(texts)
                return train_data_features, vectorizer
            else:
                return vectorizer.transform(texts), None
        elif self.representation_type == RepresentationType.WORD_EMBEDDING:
            return self.sentence_embedding_service.get_sentence_embeddings_representation(texts, language=language), None

    def get_model_dir_name(self): # for backward compatibility, we override the default get_model_dir_name()
        return "svm"


class SVM_BOW(SVM):
    def __init__(self, output_dir, background_jobs_manager, sentence_embedding_service):
        super().__init__(output_dir=output_dir, background_jobs_manager=background_jobs_manager,
                         representation_type=RepresentationType.BOW,
                         sentence_embedding_service=sentence_embedding_service)

    def get_supported_languages(self):
        return Languages.all_languages()


class SVM_WordEmbeddings(SVM):
    def __init__(self, output_dir, background_jobs_manager, sentence_embedding_service):
        super().__init__(output_dir=output_dir, background_jobs_manager=background_jobs_manager,
                         representation_type=RepresentationType.WORD_EMBEDDING,
                         sentence_embedding_service=sentence_embedding_service)

    def get_supported_languages(self):
        return Languages.all_languages()



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import torch

INFER_CACHE_SIZE = 10000000
ACTIVE_LEARNING_SUGGESTION_COUNT = 1000
CPU_WORKERS = 10  # TODO use number of cores?
MPS_GPU_AVAILABLE = hasattr(torch.backends, "mps") and torch.backends.mps.is_available()  # relevant for mac machines
# with GPU devices (e.g., Apple M1 chip). Check if mps exists in torch for backward compatibility
GPU_AVAILABLE = torch.cuda.is_available() or MPS_GPU_AVAILABLE
GPU_WORKERS = 1  # Currently only one GPU is supported













#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from label_sleuth.models.core.model_type import ModelType
from label_sleuth.models.ensemble import SVM_Ensemble
from label_sleuth.models.hf_transformers import HFBert, HFXLMRoberta
from label_sleuth.models.naive_bayes import NaiveBayes_BOW, NaiveBayes_WordEmbeddings
from label_sleuth.models.random_model import RandomModel
from label_sleuth.models.svm import SVM_BOW, SVM_WordEmbeddings


class ModelsCatalog:
    RAND = ModelType(RandomModel)
    NB_OVER_BOW = ModelType(NaiveBayes_BOW)
    NB_OVER_WORD_EMBEDDINGS = ModelType(NaiveBayes_WordEmbeddings)
    SVM_OVER_BOW = ModelType(SVM_BOW)
    SVM_OVER_WORD_EMBEDDINGS = ModelType(SVM_WordEmbeddings)
    SVM_ENSEMBLE = ModelType(SVM_Ensemble)
    HF_BERT = ModelType(HFBert)
    HF_XLM_ROBERTA = ModelType(HFXLMRoberta)




#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import os
import pickle

from dataclasses import dataclass

import numpy as np

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.naive_bayes import GaussianNB, MultinomialNB, _BaseNB

from label_sleuth.models.core.languages import Language, Languages
from label_sleuth.models.core.model_api import ModelAPI
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.models.core.tools import RepresentationType, SentenceEmbeddingService
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s')


@dataclass
class NaiveBayesModelComponents:
    model: _BaseNB
    vectorizer: None
    language: Language


class NaiveBayes(ModelAPI):
    def __init__(self, output_dir, representation_type: RepresentationType,
                 background_jobs_manager: BackgroundJobsManager,
                 sentence_embedding_service: SentenceEmbeddingService,
                 max_datapoints=10000):
        super().__init__(output_dir, background_jobs_manager)
        self.max_datapoints = max_datapoints
        self.infer_batch_size = max_datapoints
        self.representation_type = representation_type
        if self.representation_type == RepresentationType.WORD_EMBEDDING:
            self.sentence_embedding_service = sentence_embedding_service

    def _train(self, model_id, train_data, model_params):
        model = MultinomialNB() if self.representation_type == RepresentationType.BOW else GaussianNB()
        language = self.get_language(self.get_model_dir_by_id(model_id))
        texts = [x['text'] for x in train_data]
        texts = texts[:self.max_datapoints]
        train_data_features, vectorizer = self.input_to_features(texts, language=language)
        labels = [x['label'] for x in train_data]
        labels = labels[:self.max_datapoints]
        model.fit(train_data_features, labels)

        with open(os.path.join(self.get_model_dir_by_id(model_id), "vectorizer"), "wb") as fl:
            pickle.dump(vectorizer, fl)
        with open(os.path.join(self.get_model_dir_by_id(model_id), "model"), "wb") as fl:
            pickle.dump(model, fl)

    def load_model(self, model_path) -> NaiveBayesModelComponents:
        with open(os.path.join(model_path, "model"), "rb") as fl:
            model = pickle.load(fl)
        with open(os.path.join(model_path, "vectorizer"), "rb") as fl:
            vectorizer = pickle.load(fl)
        language = self.get_language(model_path)
        return NaiveBayesModelComponents(model=model, vectorizer=vectorizer, language=language)

    def infer(self, model_components: NaiveBayesModelComponents, items_to_infer):
        items_to_infer = [x['text'] for x in items_to_infer]
        last_batch = 0
        predictions = []
        while last_batch < len(items_to_infer):
            batch = items_to_infer[last_batch:last_batch + self.infer_batch_size]
            last_batch += self.infer_batch_size
            batch, _ = self.input_to_features(batch, language=model_components.language,
                                              vectorizer=model_components.vectorizer)
            predictions.append(model_components.model.predict_proba(batch))
        predictions = np.concatenate(predictions, axis=0)

        labels = [bool(np.argmax(prediction)) for prediction in predictions]
        # The True label is in the second position as sorted([True, False]) is [False, True]
        scores = [prediction[1] for prediction in predictions]
        return [Prediction(label=label, score=score) for label, score in zip(labels, scores)]

    def input_to_features(self, texts, language=Languages.ENGLISH, vectorizer=None):
        if self.representation_type == RepresentationType.BOW:
            if vectorizer is None:
                vectorizer = CountVectorizer(analyzer="word", tokenizer=None, preprocessor=None, stop_words=None,
                                             lowercase=True, max_features=1000)
                train_data_features = vectorizer.fit_transform(texts)
                return train_data_features, vectorizer
            else:
                return vectorizer.transform(texts), None
        elif self.representation_type == RepresentationType.WORD_EMBEDDING:
            return self.sentence_embedding_service.get_sentence_embeddings_representation(texts, language=language), None

    def get_model_dir_name(self): # for backward compatibility, we override the default get_model_dir_name()
        return "nb"


class NaiveBayes_BOW(NaiveBayes):
    def __init__(self, output_dir, background_jobs_manager, sentence_embedding_service):
        super().__init__(output_dir=output_dir, background_jobs_manager=background_jobs_manager,
                         representation_type=RepresentationType.BOW,
                         sentence_embedding_service=sentence_embedding_service)

    def get_supported_languages(self):
        return Languages.all_languages()



class NaiveBayes_WordEmbeddings(NaiveBayes):
    def __init__(self, output_dir, background_jobs_manager, sentence_embedding_service):
        super().__init__(output_dir=output_dir, background_jobs_manager=background_jobs_manager,
                         representation_type=RepresentationType.WORD_EMBEDDING,
                         sentence_embedding_service=sentence_embedding_service)

    def get_supported_languages(self):
        return Languages.all_languages()



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import itertools
import logging
import random
import time
from collections import defaultdict
from typing import List, Mapping, Tuple

import numpy as np
from sklearn.neighbors import NearestNeighbors

from label_sleuth.analysis_utils.analyze_tokens import get_token_overlap
from label_sleuth.data_access.core.data_structs import LABEL_POSITIVE, LABEL_NEGATIVE, TextElement, BINARY_LABELS
from label_sleuth.models.core.languages import Language
from label_sleuth.models.core.catalog import ModelsCatalog
from label_sleuth.models.core.tools import remove_stop_words_and_punctuation
from label_sleuth.orchestrator.utils import convert_text_elements_to_train_data

MIN_TOKEN_OVERLAP_THRESHOLD = 0.6


def get_disagreements_using_cross_validation(workspace_id, category_id: int, labeled_elements: List[TextElement],
                                             model_factory, language: Language, model_type=ModelsCatalog.SVM_ENSEMBLE,
                                             num_folds=4):
    """
    This method uses cross-validation in order to identify elements where the user label might be incorrect.
    The *labeled elements* are divided into *num_folds* parts. In each fold, one part is left out and the other parts
    are used to train a model; Then, the model for each fold outputs predictions for the left out elements. Where one
    of the models gives a confident prediction for an element that disagrees with the user label for that element,
    the user label is considered suspect.

    :param workspace_id:
    :param category_id:
    :param labeled_elements:
    :param model_factory:
    :param language:
    :param model_type: the type of model used for training and inference.
    :param num_folds: determines the number of partitions of the labeled elements, and thus the number of models
    that will be trained.
    :return: a list of labeled text elements, sorted from most to least "suspicious" according to the confidence of the
    model predictions
    """
    start_time = time.time()
    if {text_element.category_to_label[category_id].label for text_element in labeled_elements} != BINARY_LABELS:
        # if there are no positive labels or no negative labels, return an empty list
        return []
    model_api = model_factory.get_model_api(model_type)

    all_scores = []
    random.Random(0).shuffle(labeled_elements)
    all_train_data = convert_text_elements_to_train_data(labeled_elements, category_id)
    train_splits = np.array_split(np.array(all_train_data), num_folds)
    if any({x["label"] for x in train_splits[idx]} != BINARY_LABELS for idx in range(num_folds)):
        # if there are no positive labels or no negative labels in one of the folds, return an empty list
        return []
    for i in range(num_folds):
        # train split i is used for inference and left out of the train data
        left_out_data = train_splits[i]
        fold_train_data = \
            np.concatenate([part for j, part in enumerate(train_splits) if j != i])
        model_id, future = model_api.train(fold_train_data, language=language)
        logging.info(f'Suspicious labels report fold {i}: training cross-validation model {model_id}')
        future.result(timeout=60)
        logging.info(f'Suspicious labels report fold {i}: done waiting for cross-validation model {model_id}, '
                     f'inferring on {len(left_out_data)} left-out examples')
        fold_scores = [pred.score for pred in model_api.infer_by_id(model_id, left_out_data, use_cache=False)]
        all_scores.extend(fold_scores)
        model_api.delete_model(model_id)
    # We take elements where the fold model disagreed with the user label, but ignore model predictions that are close
    # to the decision boundary of 0.5
    disagreement_scores_and_elements = \
        [(score, element) for score, element, element_dict in zip(all_scores, labeled_elements, all_train_data)
         if score < 0.4 and element_dict['label'] == 1 or score > 0.6 and element_dict['label'] == 0]
    # sort elements by model confidence
    sorted_scores_and_elements = sorted(disagreement_scores_and_elements, key=lambda x: abs(x[0] - 0.5),
                                        reverse=True)
    sorted_disagreement_elements = [text_element for score, text_element in sorted_scores_and_elements]
    logging.info(f"creating suspicious labels report for category_id '{category_id}' in workspace '{workspace_id}' "
                 f"took {'{:.2f}'.format(time.time() - start_time)} seconds")
    return sorted_disagreement_elements


def get_suspected_labeling_contradictions_by_distance_with_diffs(category_id: int, labeled_elements, embedding_func,
                                                                 language: Language) -> Mapping[str, List]:
    """
    Enrich the output of possibly inconsistent element pairs from *get_suspected_labeling_contradictions_by_distance*
    with sets of tokens that differentiate the pair of elements, to enable highlighting the similarities/differences
    between the texts in each pair.

    :param category_id:
    :param labeled_elements:
    :param embedding_func: a function that receives a list of texts and a language, and returns a list of embedding
    vectors for those texts.
    :param language:
    :return: a dictionary containing:
    {'pairs': the list of text element pairs from get_suspected_labeling_contradictions_by_distance,
    i.e. [[pair_1_element_a, pair_1_element_b], [pair_2_element_a, pair_2_element_b], ...],
    'diffs': a list of tuples with the tokens unique to the first and the second element of each pair, respectively,
    i.e. [(pair_1_element_a_unique_tokens_set, pair_1_element_b_unique_tokens_set), ...]
    """
    pairs = get_suspected_labeling_contradictions_by_distance(category_id, labeled_elements, embedding_func, language)
    diffs = []
    for pair in pairs:
        set1 = set(pair[0].text.split())
        set2 = set(pair[1].text.split())
        intersect = set1.intersection(set2)
        diffs.append((set1 - intersect, set2 - intersect))
    return {'pairs': pairs, 'diffs': diffs}


def get_suspected_labeling_contradictions_by_distance(category_id, labeled_elements: List[TextElement],
                                                      embedding_func, language: Language) \
        -> List[List[TextElement]]:
    """
    This method uses text embeddings in order to identify user labels that may be inconsistent with each other.
    For each of the *labeled_elements*, we identify a nearest neighbor that was given the opposite label.
    Where the similarity between the pair of opposite-label elements is high, we suspect that one of the pair elements
    may have been given the wrong label.

    :param category_id:
    :param labeled_elements:
    :param embedding_func: a function that receives a list of texts and a language, and returns a list of embedding
    vectors for those texts.
    :param language:
    :return: a list of text element pairs, sorted from high to low similarity between the pair text embeddings.
    """
    if {text_element.category_to_label[category_id].label for text_element in labeled_elements} != BINARY_LABELS:
        # if there are no positive labels or no negative labels, return an empty list
        return []

    embedding_vectors = embedding_func([e.text for e in labeled_elements], language=language)
    sorted_pair_lists = []
    # for each positively labeled element we find the nearest neighbor with a negative label, and vice versa
    for source_label in [LABEL_POSITIVE, LABEL_NEGATIVE]:
        distances_and_pairs = _get_nearest_neighbors_with_opposite_label(labeled_elements, embedding_vectors,
                                                                         category_id, source_label=source_label)
        sorted_element_pairs = [element_pair for distance, element_pair
                                in sorted(distances_and_pairs, key=lambda x: x[0])]  # sort by distance
        sorted_pair_lists.append(sorted_element_pairs)

    # interleave positive-to-negative and negative-to-positive neighbor pairs
    unified_pairs_list = [tup for subset in itertools.zip_longest(*sorted_pair_lists) for tup in subset
                          if tup is not None]
    # filter duplicates and low-overlap pairs
    unified_pairs_list = _filter_nearest_neighbor_pairs(unified_pairs_list, language=language)
    # align pairs by label
    unified_pairs_list = [sorted(pair, key=lambda te: te.category_to_label[category_id].label, reverse=True)
                          for pair in unified_pairs_list]
    return unified_pairs_list


def _get_nearest_neighbors_with_opposite_label(all_elements: List[TextElement], embedding_vectors: List, category_id,
                                               source_label) -> List[Tuple[float, Tuple[TextElement, TextElement]]]:
    source_label_idxs = [i for i, (element, rep) in enumerate(zip(all_elements, embedding_vectors))
                         if element.category_to_label[category_id].label == source_label and rep[0] != 0]
    opposite_label_idxs = [i for i, (element, rep) in enumerate(zip(all_elements, embedding_vectors))
                           if element.category_to_label[category_id].label != source_label and rep[0] != 0]
    source_label_embeddings = np.array(embedding_vectors)[source_label_idxs]
    opposite_label_embeddings = np.array(embedding_vectors)[opposite_label_idxs]

    nbrs = NearestNeighbors(n_neighbors=1, algorithm='ball_tree').fit(opposite_label_embeddings)
    distances_to_closest_opposite, indices_of_closest_opposite = nbrs.kneighbors(source_label_embeddings)

    distances_to_closest_opposite = [d[0] for d in distances_to_closest_opposite] # one neighbor for each element
    indices_of_closest_opposite = [d[0] for d in indices_of_closest_opposite]

    distances_and_pairs = \
        [(distance,
          (all_elements[source_label_idxs[i]], all_elements[opposite_label_idxs[opposite_neighbor_idx]]))
         for i, (distance, opposite_neighbor_idx) in enumerate(zip(distances_to_closest_opposite,
                                                                   indices_of_closest_opposite))]
    return distances_and_pairs


def _filter_nearest_neighbor_pairs(pairs_list: List[Tuple[TextElement]], language: Language):
    """
    :param pairs_list: a list of tuples of TextElements, where the element in tuple index 1 is the nearest neighbor of
    the element in tuple index 0
    :param language:
    :return: a subset of pairs_list, preserving the order from the original list
    """
    # remove duplicate pairs
    unique_pairs = set()
    filtered_pairs_list = [(element_a, element_b) for element_a, element_b in pairs_list
                           if frozenset({element_a.uri, element_b.uri}) not in unique_pairs
                           and not unique_pairs.add(frozenset({element_a.uri, element_b.uri}))]

    # remove pairs with low token overlap
    filtered_pairs_list = \
        [(element_a, element_b) for element_a, element_b in filtered_pairs_list
         if get_token_overlap(*remove_stop_words_and_punctuation([element_a.text, element_b.text], language=language))
         > MIN_TOKEN_OVERLAP_THRESHOLD]

    # remove elements that appear too many times (currently this is a maximum of 3 appearances: each element can appear
    # once as a source element, and 0-2 times as a nearest neighbor to another source element)
    uri_counts = defaultdict(lambda: 0)
    filtered_pairs_list = [(source_element, neighbor) for source_element, neighbor in filtered_pairs_list
                           if uri_counts[neighbor.uri] <= 1 and not uri_counts.update({neighbor.uri:
                                                                                           uri_counts[
                                                                                               neighbor.uri] + 1})]
    return filtered_pairs_list



import os
import tempfile
import unittest
import pandas as pd

from label_sleuth.data_access.processors.csv_processor import CsvFileProcessor


class TestCsvFileProcessor(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()

    def test_csv_processor_with_document_id(self):
        workspace_id = self.test_csv_processor_with_document_id.__name__
        dataset_name = f'{workspace_id}_dump'
        data = [{"document_id": 0, "text": "first sentence in document 0"},
                {"document_id": 0, "text": "second sentence in document 0"},
                {"document_id": 1, "text": "first sentence in document 1"},
                {"document_id": 2, "text": "first sentence in document 2", "metadata_text_subject": "no subject"},
                {"document_id": 2, "text": "second sentence in document 2"},
                {"document_id": 2, "text": "third sentence in document 2"}]
        df = pd.DataFrame(data)
        file_path = os.path.join(TestCsvFileProcessor.temp_dir.name, "dataset_name.csv")
        df.to_csv(file_path)
        processor = CsvFileProcessor(dataset_name, file_path)
        documents = processor.build_documents()
        self.assertEqual(3, len(documents))

        # assert URIs are equal
        for sentence in data:
            self.assertEqual(f"{dataset_name}-{sentence['document_id']}", documents[sentence['document_id']].uri)

        # assert text is equal
        all_texts_from_documents = [text_element for document in documents for text_element in document.text_elements]
        for expected_dict, text_element in zip(data, all_texts_from_documents):
            self.assertEqual(expected_dict["text"], text_element.text)

        # assert metadata is equal
        self.assertEqual({"text_subject": "no subject"}, documents[2].text_elements[0].metadata)

    def test_csv_processor_without_document_id(self):
        workspace_id = self.test_csv_processor_without_document_id.__name__
        dataset_name = f'{workspace_id}_dump'
        data = [{"text": "1st sentence"},
                {"text": "2nd sentence"},
                {"text": "3rd sentence"},
                {"text": "4th sentence"},
                {"text": "5th sentence"},
                {"text": "6th sentence"},
                {"text": "7th sentence"},
                {"text": "8th sentence"},
                {"text": "9th sentence"},
                {"text": "10th sentence"},
                {"text": "11th sentence"}]
        df = pd.DataFrame(data)
        file_path = os.path.join(TestCsvFileProcessor.temp_dir.name, "dataset_name.csv")
        df.to_csv(file_path)
        processor = CsvFileProcessor(dataset_name, file_path)
        documents = processor.build_documents()
        self.assertEqual(1, len(documents))

        # assert text is equal
        all_texts_from_documents = documents[0].text_elements
        for expected_dict, text_element in zip(data, all_texts_from_documents):
            self.assertTrue(text_element.uri.startswith(dataset_name))
            self.assertEqual(expected_dict["text"], text_element.text)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import re
import os
import shutil
import string

import tarfile
import tempfile
import threading
import fasttext
import fasttext.util

from collections import defaultdict
from enum import Enum
from typing import List

import numpy as np
import requests
import spacy
from tqdm.auto import tqdm

from label_sleuth.models.core.languages import Language, Languages


class RepresentationType(Enum):
    WORD_EMBEDDING = 1
    BOW = 2


class SentenceEmbeddingService:
    def __init__(self, embedding_model_dir, preload_spacy_model_name=None, preload_fasttext_language_id=None):
        self.spacy_models_path = os.path.join(embedding_model_dir, "spacy_models")
        self.fasttext_models_path = os.path.join(embedding_model_dir, "fasttext_models")
        fasttext.FastText.eprint = lambda x: None
        os.makedirs(self.spacy_models_path, exist_ok=True)
        os.makedirs(self.fasttext_models_path, exist_ok=True)
        self.spacy_models = defaultdict(lambda: None)
        self.spacy_model_lock = threading.Lock()
        if preload_spacy_model_name is not None:
            self.load_or_download_spacy_model(preload_spacy_model_name)
        if preload_fasttext_language_id is not None:
            self.get_fasttext_model(preload_fasttext_language_id)

    def get_sentence_embeddings_representation(self, sentences: List[str],
                                               language: Language = Languages.ENGLISH) -> List[np.ndarray]:
        """
        Given a list of texts, return a list of representation vectors. Each text is represented by the
        mean of the vector representations of its consisting tokens.
        :param sentences:
        :param language:
        :return: a list of numpy vectors. Vector length depends on the representation model specified under *language*
        """

        sentences = remove_stop_words_and_punctuation(sentences, language=language)

        if language.spacy_model_name is not None:
            model_name = language.spacy_model_name

            # The model used for calculating the representations
            spacy_model = self.get_spacy_model(model_name)

            # remove out-of-vocabulary tokens
            sentences = [' '.join(token for token in sent.split() if spacy_model.vocab.has_vector(token))
                         for sent in sentences]
            # the vector obtained by *make_doc(X).vector* is an average of the representations for the
            # individual tokens in X
            embeddings = [spacy_model.make_doc(sent).vector for sent in sentences]
            logging.info(f"Done getting GloVe representations for {len(embeddings)} sentences")
        else:
            # fasttext model
            model = self.get_fasttext_model(language.fasttext_language_id)
            embeddings = [model.get_sentence_vector(sent) for sent in sentences]
            logging.info(f"Done getting FastText representations for {len(embeddings)} sentences")

        return embeddings

    def get_spacy_model(self, model_name) -> spacy.Language:
        """
        The model is loaded once, on the first time this method is called. On
        subsequent calls, the loaded model is read from the spacy_models dictionary
        """
        with self.spacy_model_lock:
            if self.spacy_models[model_name] is None:
                self.spacy_models[model_name] = self.load_or_download_spacy_model(model_name)
        return self.spacy_models[model_name]

    @staticmethod
    def download_spacy_model(model_name, output_path):
        """
        The default spacy download method installs spacy models via pip install, which is unsuitable for some
        deployment and packaging scenarios. Thus, in this method we download the spacy model archive manually, and
        then place the model files inside the Label Sleuth output directory.
        See https://spacy.io/usage/models#download-manual

        """
        from spacy.cli._util import SDIST_SUFFIX
        from spacy.cli.download import get_compatibility, get_version

        # get the url for the relevant version of the requested spacy model
        compatibility = get_compatibility()
        model_version = get_version(model_name, compatibility)
        model_with_version = f'{model_name}-{model_version}'
        full_download_url = f'{spacy.about.__download_url__}/{model_with_version}/{model_with_version}{SDIST_SUFFIX}'

        # download the archive file
        temp_path = tempfile.TemporaryDirectory().name
        with requests.get(full_download_url, stream=True) as response:
            total_length = int(response.headers.get("Content-Length"))
            with tqdm.wrapattr(response.raw, "read", total=total_length, desc="Downloading spacy model") as raw:
                with open(temp_path, 'wb') as output:
                    shutil.copyfileobj(raw, output)

        tar = tarfile.open(temp_path)
        temp_path_2 = tempfile.TemporaryDirectory().name
        # extract the relevant directory from the archive (https://spacy.io/usage/models#download-manual)
        path_to_model_files = f'{model_with_version}/{model_name}/{model_with_version}'
        tar.extractall(members=[x for x in tarfile.open(temp_path).getmembers()
                                if x.name.startswith(path_to_model_files)],
                       path=temp_path_2)

        # move the model files to the Label Sleuth output directory
        shutil.move(os.path.join(temp_path_2, path_to_model_files), output_path)

    def load_or_download_spacy_model(self, model_name) -> spacy.Language:
        """
        Load or download a spacy model by name, and place it inside self.spacy_models_path
        """
        model_path = os.path.join(self.spacy_models_path, model_name)
        if os.path.exists(model_path):
            return spacy.load(model_path)

        logging.info(f"Spacy model does not exist in {model_path}, downloading...")
        self.download_spacy_model(model_name, model_path)
        model = spacy.load(model_path)
        return model

    def get_fasttext_model(self, fasttext_language_id):
        model_file_name = f"cc.{fasttext_language_id}.300.bin"
        model_path = os.path.join(self.fasttext_models_path, model_file_name)
        if os.path.exists(model_path):
            return fasttext.load_model(model_path)
        logging.info(f"fasttext model for language code '{fasttext_language_id}' does not exist in "
                     f"{model_path}, downloading...")
        original_cwd = os.getcwd()

        try:
            os.chdir(self.fasttext_models_path)
            fasttext.util.download_model(fasttext_language_id, if_exists='ignore')
            os.remove(os.path.join(self.fasttext_models_path, f"{model_file_name}.gz"))
        except Exception as e:
            raise Exception(f"Could not download FastText model for language id {fasttext_language_id}") from e
        finally:
            os.chdir(original_cwd)
        return fasttext.load_model(model_path)


def remove_stop_words_and_punctuation(sentences: List[str], language=Languages.ENGLISH) -> List[str]:
    # remove punctuation
    sentences = remove_punctuation(sentences)
    # remove stop words
    regex = r"\b(" + "|".join(language.stop_words) + r")\b"
    sentences = [re.sub(regex, r"", text) for text in sentences]
    # remove extra spaces
    sentences = [' '.join(sent.split()) for sent in sentences]
    return sentences


def remove_punctuation(sentences: List[str]) -> List[str]:
    punctuation = string.punctuation + '•●'
    sentences = [t.translate(t.maketrans(punctuation, ' ' * len(punctuation))) for t in sentences]
    return sentences



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import random
import tempfile
import unittest

from label_sleuth.data_access.core.data_structs import LABEL_POSITIVE, LABEL_NEGATIVE
from label_sleuth.models.core.catalog import ModelsCatalog
from label_sleuth.models.core.languages import Languages
from label_sleuth.models.core.model_api import ModelStatus
from label_sleuth.models.core.models_factory import ModelFactory
from label_sleuth.models.ensemble import Ensemble
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

PREFIX = 'Fascinating sentence'


def train_failed_with_error(*args, **kwargs):
    raise Exception("Training failed")


class TestEnsembleModel(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.background_job_manager = BackgroundJobsManager()
        model_factory = ModelFactory(self.temp_dir.name, self.background_job_manager, None)
        self.ensemble = Ensemble(self.temp_dir.name, model_types=[ModelsCatalog.RAND, ModelsCatalog.NB_OVER_BOW],
                                 background_jobs_manager=self.background_job_manager,
                                 model_factory=model_factory)
        all_nums = list(range(1, 15))
        random.shuffle(all_nums)
        self.sentences = [{'text': f'{PREFIX} {num}', 'label': random.choice([LABEL_POSITIVE, LABEL_NEGATIVE])}
                          for num in all_nums]

    def test_one_ensemble_model_fails(self):
        self.ensemble.model_apis[1]._train = train_failed_with_error
        ensemble_model_id, future = self.ensemble.train(self.sentences, Languages.ENGLISH)
        self.assertRaises(Exception, future.result)

    def test_ensemble_flow(self):
        ensemble_model_id, future = self.ensemble.train(self.sentences, Languages.ENGLISH)

        model_ids = ensemble_model_id.split(',')
        self.assertEqual(len(model_ids), len(self.ensemble.model_apis))

        future.result()
        for model_id, model_api in zip(model_ids, self.ensemble.model_apis):
            self.assertEqual(model_api.get_model_status(model_id), ModelStatus.READY)

        ensemble_predictions = self.ensemble.infer_by_id(ensemble_model_id, self.sentences)
        for ensemble_pred in ensemble_predictions:
            scores = [ensemble_pred.model_type_to_prediction[t.name].score for t in self.ensemble.model_types]
            self.assertEqual(ensemble_pred.score, self.ensemble.aggregation_func(scores))

    #TODO add a test for ensemble model path validation once export is fully implemented

    def tearDown(self):
        self.temp_dir.cleanup()
